@@ -1,5 +1,7 @@
 #[cfg(unix)]
 use std::path::Path;
+#[cfg(feature = "tls-rustls")]
+use std::sync::Arc;
 use std::{
     future::Future,
     io,
@@ -10,8 +12,15 @@ use std::{
 
 use crate::aio::{AsyncStream, RedisRuntime};
 use crate::types::RedisResult;
-#[cfg(feature = "tls")]
+
+#[cfg(all(feature = "tls-native-tls", not(feature = "tls-rustls")))]
 use async_native_tls::{TlsConnector, TlsStream};
+
+#[cfg(feature = "tls-rustls")]
+use crate::connection::create_rustls_config;
+#[cfg(feature = "tls-rustls")]
+use futures_rustls::{client::TlsStream, TlsConnector};
+
 use async_std::net::TcpStream;
 #[cfg(unix)]
 use async_std::os::unix::net::UnixStream;
@@ -19,6 +28,33 @@ use async_trait::async_trait;
 use futures_util::ready;
 use tokio::io::{AsyncRead, AsyncWrite, ReadBuf};
 
+#[inline(always)]
+async fn connect_tcp(addr: &SocketAddr) -> io::Result<TcpStream> {
+    let socket = TcpStream::connect(addr).await?;
+    #[cfg(feature = "tcp_nodelay")]
+    socket.set_nodelay(true)?;
+    #[cfg(feature = "keep-alive")]
+    {
+        //For now rely on system defaults
+        const KEEP_ALIVE: socket2::TcpKeepalive = socket2::TcpKeepalive::new();
+        //these are useless error that not going to happen
+        let mut std_socket = std::net::TcpStream::try_from(socket)?;
+        let socket2: socket2::Socket = std_socket.into();
+        socket2.set_tcp_keepalive(&KEEP_ALIVE)?;
+        std_socket = socket2.into();
+        Ok(std_socket.into())
+    }
+    #[cfg(not(feature = "keep-alive"))]
+    {
+        Ok(socket)
+    }
+}
+#[cfg(feature = "tls-rustls")]
+use crate::tls::TlsConnParams;
+
+#[cfg(all(feature = "tls-native-tls", not(feature = "tls-rustls")))]
+use crate::connection::TlsConnParams;
+
 pin_project_lite::pin_project! {
     /// Wraps the async_std `AsyncRead/AsyncWrite` in order to implement the required the tokio traits
     /// for it
@@ -82,8 +118,11 @@ pub enum AsyncStd {
     /// Represents an Async_std TCP connection.
     Tcp(AsyncStdWrapped<TcpStream>),
     /// Represents an Async_std TLS encrypted TCP connection.
-    #[cfg(feature = "async-std-tls-comp")]
-    TcpTls(AsyncStdWrapped<TlsStream<TcpStream>>),
+    #[cfg(any(
+        feature = "async-std-native-tls-comp",
+        feature = "async-std-rustls-comp"
+    ))]
+    TcpTls(AsyncStdWrapped<Box<TlsStream<TcpStream>>>),
     /// Represents an Async_std Unix connection.
     #[cfg(unix)]
     Unix(AsyncStdWrapped<UnixStream>),
@@ -97,7 +136,10 @@ impl AsyncWrite for AsyncStd {
     ) -> Poll<io::Result<usize>> {
         match &mut *self {
             AsyncStd::Tcp(r) => Pin::new(r).poll_write(cx, buf),
-            #[cfg(feature = "async-std-tls-comp")]
+            #[cfg(any(
+                feature = "async-std-native-tls-comp",
+                feature = "async-std-rustls-comp"
+            ))]
             AsyncStd::TcpTls(r) => Pin::new(r).poll_write(cx, buf),
             #[cfg(unix)]
             AsyncStd::Unix(r) => Pin::new(r).poll_write(cx, buf),
@@ -107,7 +149,10 @@ impl AsyncWrite for AsyncStd {
     fn poll_flush(mut self: Pin<&mut Self>, cx: &mut task::Context) -> Poll<io::Result<()>> {
         match &mut *self {
             AsyncStd::Tcp(r) => Pin::new(r).poll_flush(cx),
-            #[cfg(feature = "async-std-tls-comp")]
+            #[cfg(any(
+                feature = "async-std-native-tls-comp",
+                feature = "async-std-rustls-comp"
+            ))]
             AsyncStd::TcpTls(r) => Pin::new(r).poll_flush(cx),
             #[cfg(unix)]
             AsyncStd::Unix(r) => Pin::new(r).poll_flush(cx),
@@ -117,7 +162,10 @@ impl AsyncWrite for AsyncStd {
     fn poll_shutdown(mut self: Pin<&mut Self>, cx: &mut task::Context) -> Poll<io::Result<()>> {
         match &mut *self {
             AsyncStd::Tcp(r) => Pin::new(r).poll_shutdown(cx),
-            #[cfg(feature = "async-std-tls-comp")]
+            #[cfg(any(
+                feature = "async-std-native-tls-comp",
+                feature = "async-std-rustls-comp"
+            ))]
             AsyncStd::TcpTls(r) => Pin::new(r).poll_shutdown(cx),
             #[cfg(unix)]
             AsyncStd::Unix(r) => Pin::new(r).poll_shutdown(cx),
@@ -133,7 +181,10 @@ impl AsyncRead for AsyncStd {
     ) -> Poll<io::Result<()>> {
         match &mut *self {
             AsyncStd::Tcp(r) => Pin::new(r).poll_read(cx, buf),
-            #[cfg(feature = "async-std-tls-comp")]
+            #[cfg(any(
+                feature = "async-std-native-tls-comp",
+                feature = "async-std-rustls-comp"
+            ))]
             AsyncStd::TcpTls(r) => Pin::new(r).poll_read(cx, buf),
             #[cfg(unix)]
             AsyncStd::Unix(r) => Pin::new(r).poll_read(cx, buf),
@@ -144,18 +195,19 @@ impl AsyncRead for AsyncStd {
 #[async_trait]
 impl RedisRuntime for AsyncStd {
     async fn connect_tcp(socket_addr: SocketAddr) -> RedisResult<Self> {
-        Ok(TcpStream::connect(&socket_addr)
+        Ok(connect_tcp(&socket_addr)
             .await
             .map(|con| Self::Tcp(AsyncStdWrapped::new(con)))?)
     }
 
-    #[cfg(feature = "tls")]
+    #[cfg(all(feature = "tls-native-tls", not(feature = "tls-rustls")))]
     async fn connect_tcp_tls(
         hostname: &str,
         socket_addr: SocketAddr,
         insecure: bool,
+        _tls_params: &Option<TlsConnParams>,
     ) -> RedisResult<Self> {
-        let tcp_stream = TcpStream::connect(&socket_addr).await?;
+        let tcp_stream = connect_tcp(&socket_addr).await?;
         let tls_connector = if insecure {
             TlsConnector::new()
                 .danger_accept_invalid_certs(true)
@@ -167,7 +219,28 @@ impl RedisRuntime for AsyncStd {
         Ok(tls_connector
             .connect(hostname, tcp_stream)
             .await
-            .map(|con| Self::TcpTls(AsyncStdWrapped::new(con)))?)
+            .map(|con| Self::TcpTls(AsyncStdWrapped::new(Box::new(con))))?)
+    }
+
+    #[cfg(feature = "tls-rustls")]
+    async fn connect_tcp_tls(
+        hostname: &str,
+        socket_addr: SocketAddr,
+        insecure: bool,
+        tls_params: &Option<TlsConnParams>,
+    ) -> RedisResult<Self> {
+        let tcp_stream = connect_tcp(&socket_addr).await?;
+
+        let config = create_rustls_config(insecure, tls_params.clone())?;
+        let tls_connector = TlsConnector::from(Arc::new(config));
+
+        Ok(tls_connector
+            .connect(
+                rustls_pki_types::ServerName::try_from(hostname)?.to_owned(),
+                tcp_stream,
+            )
+            .await
+            .map(|con| Self::TcpTls(AsyncStdWrapped::new(Box::new(con))))?)
     }
 
     #[cfg(unix)]
@@ -184,7 +257,10 @@ impl RedisRuntime for AsyncStd {
     fn boxed(self) -> Pin<Box<dyn AsyncStream + Send + Sync>> {
         match self {
             AsyncStd::Tcp(x) => Box::pin(x),
-            #[cfg(feature = "async-std-tls-comp")]
+            #[cfg(any(
+                feature = "async-std-native-tls-comp",
+                feature = "async-std-rustls-comp"
+            ))]
             AsyncStd::TcpTls(x) => Box::pin(x),
             #[cfg(unix)]
             AsyncStd::Unix(x) => Box::pin(x),
