@@ -1,7 +1,8 @@
 #[cfg(feature = "aio")]
 use futures_util::{
+    future::BoxFuture,
     task::{Context, Poll},
-    FutureExt, Stream,
+    Stream, StreamExt,
 };
 #[cfg(feature = "aio")]
 use std::pin::Pin;
@@ -9,7 +10,7 @@ use std::{fmt, io};
 
 use crate::connection::ConnectionLike;
 use crate::pipeline::Pipeline;
-use crate::types::{from_redis_value, FromRedisValue, RedisResult, RedisWrite, ToRedisArgs};
+use crate::types::{from_owned_redis_value, FromRedisValue, RedisResult, RedisWrite, ToRedisArgs};
 
 /// An argument to a redis command
 #[derive(Clone)]
@@ -27,6 +28,8 @@ pub struct Cmd {
     // Arg::Simple contains the offset that marks the end of the argument
     args: Vec<Arg<usize>>,
     cursor: Option<u64>,
+    // If it's true command's response won't be read from socket. Useful for Pub/Sub.
+    no_response: bool,
 }
 
 /// Represents a redis iterator.
@@ -54,12 +57,9 @@ impl<'a, T: FromRedisValue> Iterator for Iter<'a, T> {
                 return None;
             }
 
-            let pcmd = unwrap_or!(
-                self.cmd.get_packed_command_with_cursor(self.cursor),
-                return None
-            );
-            let rv = unwrap_or!(self.con.req_packed_command(&pcmd).ok(), return None);
-            let (cur, batch): (u64, Vec<T>) = unwrap_or!(from_redis_value(&rv).ok(), return None);
+            let pcmd = self.cmd.get_packed_command_with_cursor(self.cursor)?;
+            let rv = self.con.req_packed_command(&pcmd).ok()?;
+            let (cur, batch): (u64, Vec<T>) = from_owned_redis_value(rv).ok()?;
 
             self.cursor = cur;
             self.batch = batch.into_iter();
@@ -70,30 +70,30 @@ impl<'a, T: FromRedisValue> Iterator for Iter<'a, T> {
 #[cfg(feature = "aio")]
 use crate::aio::ConnectionLike as AsyncConnection;
 
-/// Represents a redis iterator that can be used with async connections.
+/// The inner future of AsyncIter
 #[cfg(feature = "aio")]
-pub struct AsyncIter<'a, T: FromRedisValue + 'a> {
+struct AsyncIterInner<'a, T: FromRedisValue + 'a> {
     batch: std::vec::IntoIter<T>,
     con: &'a mut (dyn AsyncConnection + Send + 'a),
     cmd: Cmd,
 }
 
+/// Represents the state of AsyncIter
 #[cfg(feature = "aio")]
-impl<'a, T: FromRedisValue + 'a> AsyncIter<'a, T> {
-    /// ```rust,no_run
-    /// # use redis::AsyncCommands;
-    /// # async fn scan_set() -> redis::RedisResult<()> {
-    /// # let client = redis::Client::open("redis://127.0.0.1/")?;
-    /// # let mut con = client.get_async_connection().await?;
-    /// con.sadd("my_set", 42i32).await?;
-    /// con.sadd("my_set", 43i32).await?;
-    /// let mut iter: redis::AsyncIter<i32> = con.sscan("my_set").await?;
-    /// while let Some(element) = iter.next_item().await {
-    ///     assert!(element == 42 || element == 43);
-    /// }
-    /// # Ok(())
-    /// # }
-    /// ```
+enum IterOrFuture<'a, T: FromRedisValue + 'a> {
+    Iter(AsyncIterInner<'a, T>),
+    Future(BoxFuture<'a, (AsyncIterInner<'a, T>, Option<T>)>),
+    Empty,
+}
+
+/// Represents a redis iterator that can be used with async connections.
+#[cfg(feature = "aio")]
+pub struct AsyncIter<'a, T: FromRedisValue + 'a> {
+    inner: IterOrFuture<'a, T>,
+}
+
+#[cfg(feature = "aio")]
+impl<'a, T: FromRedisValue + 'a> AsyncIterInner<'a, T> {
     #[inline]
     pub async fn next_item(&mut self) -> Option<T> {
         // we need to do this in a loop until we produce at least one item
@@ -112,11 +112,8 @@ impl<'a, T: FromRedisValue + 'a> AsyncIter<'a, T> {
                 return None;
             }
 
-            let rv = unwrap_or!(
-                self.con.req_packed_command(&self.cmd).await.ok(),
-                return None
-            );
-            let (cur, batch): (u64, Vec<T>) = unwrap_or!(from_redis_value(&rv).ok(), return None);
+            let rv = self.con.req_packed_command(&self.cmd).await.ok()?;
+            let (cur, batch): (u64, Vec<T>) = from_owned_redis_value(rv).ok()?;
 
             self.cmd.cursor = Some(cur);
             self.batch = batch.into_iter();
@@ -125,13 +122,55 @@ impl<'a, T: FromRedisValue + 'a> AsyncIter<'a, T> {
 }
 
 #[cfg(feature = "aio")]
-impl<'a, T: FromRedisValue + Unpin + 'a> Stream for AsyncIter<'a, T> {
+impl<'a, T: FromRedisValue + 'a + Unpin + Send> AsyncIter<'a, T> {
+    /// ```rust,no_run
+    /// # use redis::AsyncCommands;
+    /// # async fn scan_set() -> redis::RedisResult<()> {
+    /// # let client = redis::Client::open("redis://127.0.0.1/")?;
+    /// # let mut con = client.get_multiplexed_async_connection().await?;
+    /// let _: () = con.sadd("my_set", 42i32).await?;
+    /// let _: () = con.sadd("my_set", 43i32).await?;
+    /// let mut iter: redis::AsyncIter<i32> = con.sscan("my_set").await?;
+    /// while let Some(element) = iter.next_item().await {
+    ///     assert!(element == 42 || element == 43);
+    /// }
+    /// # Ok(())
+    /// # }
+    /// ```
+    #[inline]
+    pub async fn next_item(&mut self) -> Option<T> {
+        StreamExt::next(self).await
+    }
+}
+
+#[cfg(feature = "aio")]
+impl<'a, T: FromRedisValue + Unpin + Send + 'a> Stream for AsyncIter<'a, T> {
     type Item = T;
 
     fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<T>> {
         let this = self.get_mut();
-        let mut future = Box::pin(this.next_item());
-        future.poll_unpin(cx)
+        let inner = std::mem::replace(&mut this.inner, IterOrFuture::Empty);
+        match inner {
+            IterOrFuture::Iter(mut iter) => {
+                let fut = async move {
+                    let next_item = iter.next_item().await;
+                    (iter, next_item)
+                };
+                this.inner = IterOrFuture::Future(Box::pin(fut));
+                Pin::new(this).poll_next(cx)
+            }
+            IterOrFuture::Future(mut fut) => match fut.as_mut().poll(cx) {
+                Poll::Pending => {
+                    this.inner = IterOrFuture::Future(fut);
+                    Poll::Pending
+                }
+                Poll::Ready((iter, value)) => {
+                    this.inner = IterOrFuture::Iter(iter);
+                    Poll::Ready(value)
+                }
+            },
+            IterOrFuture::Empty => unreachable!(),
+        }
     }
 }
 
@@ -203,8 +242,11 @@ fn write_command<'a, I>(cmd: &mut (impl ?Sized + io::Write), args: I, cursor: u6
 where
     I: IntoIterator<Item = Arg<&'a [u8]>> + Clone + ExactSizeIterator,
 {
+    let mut buf = ::itoa::Buffer::new();
+
     cmd.write_all(b"*")?;
-    ::itoa::write(&mut *cmd, args.len())?;
+    let s = buf.format(args.len());
+    cmd.write_all(s.as_bytes())?;
     cmd.write_all(b"\r\n")?;
 
     let mut cursor_bytes = itoa::Buffer::new();
@@ -215,7 +257,8 @@ where
         };
 
         cmd.write_all(b"$")?;
-        ::itoa::write(&mut *cmd, bytes.len())?;
+        let s = buf.format(bytes.len());
+        cmd.write_all(s.as_bytes())?;
         cmd.write_all(b"\r\n")?;
 
         cmd.write_all(bytes)?;
@@ -232,7 +275,7 @@ impl RedisWrite for Cmd {
 
     fn write_arg_fmt(&mut self, arg: impl fmt::Display) {
         use std::io::Write;
-        write!(self.data, "{}", arg).unwrap();
+        write!(self.data, "{arg}").unwrap();
         self.args.push(Arg::Simple(self.data.len()));
     }
 }
@@ -244,7 +287,7 @@ impl Default for Cmd {
 }
 
 /// A command acts as a builder interface to creating encoded redis
-/// requests.  This allows you to easiy assemble a packed command
+/// requests.  This allows you to easily assemble a packed command
 /// by chaining arguments together.
 ///
 /// Basic example:
@@ -260,7 +303,7 @@ impl Default for Cmd {
 /// redis::cmd("SET").arg("my_key").arg(42);
 /// ```
 ///
-/// Because currently rust's currently does not have an ideal system
+/// Because Rust currently does not have an ideal system
 /// for lifetimes of temporaries, sometimes you need to hold on to
 /// the initially generated command:
 ///
@@ -277,9 +320,27 @@ impl Cmd {
             data: vec![],
             args: vec![],
             cursor: None,
+            no_response: false,
         }
     }
 
+    /// Creates a new empty command, with at least the requested capacity.
+    pub fn with_capacity(arg_count: usize, size_of_data: usize) -> Cmd {
+        Cmd {
+            data: Vec::with_capacity(size_of_data),
+            args: Vec::with_capacity(arg_count),
+            cursor: None,
+            no_response: false,
+        }
+    }
+
+    /// Get the capacities for the internal buffers.
+    #[cfg(test)]
+    #[allow(dead_code)]
+    pub(crate) fn capacity(&self) -> (usize, usize) {
+        (self.args.capacity(), self.data.capacity())
+    }
+
     /// Appends an argument to the command.  The argument passed must
     /// be a type that implements `ToRedisArgs`.  Most primitive types as
     /// well as vectors of primitive types implement it.
@@ -362,7 +423,7 @@ impl Cmd {
     #[inline]
     pub fn query<T: FromRedisValue>(&self, con: &mut dyn ConnectionLike) -> RedisResult<T> {
         match con.req_command(self) {
-            Ok(val) => from_redis_value(&val),
+            Ok(val) => from_owned_redis_value(val.extract_error()?),
             Err(e) => Err(e),
         }
     }
@@ -370,12 +431,12 @@ impl Cmd {
     /// Async version of `query`.
     #[inline]
     #[cfg(feature = "aio")]
-    pub async fn query_async<C, T: FromRedisValue>(&self, con: &mut C) -> RedisResult<T>
-    where
-        C: crate::aio::ConnectionLike,
-    {
+    pub async fn query_async<T: FromRedisValue>(
+        &self,
+        con: &mut impl crate::aio::ConnectionLike,
+    ) -> RedisResult<T> {
         let val = con.req_packed_command(self).await?;
-        from_redis_value(&val)
+        from_owned_redis_value(val.extract_error()?)
     }
 
     /// Similar to `query()` but returns an iterator over the items of the
@@ -397,9 +458,9 @@ impl Cmd {
         let rv = con.req_command(&self)?;
 
         let (cursor, batch) = if rv.looks_like_cursor() {
-            from_redis_value::<(u64, Vec<T>)>(&rv)?
+            from_owned_redis_value::<(u64, Vec<T>)>(rv)?
         } else {
-            (0, from_redis_value(&rv)?)
+            (0, from_owned_redis_value(rv)?)
         };
 
         Ok(Iter {
@@ -412,7 +473,7 @@ impl Cmd {
 
     /// Similar to `iter()` but returns an AsyncIter over the items of the
     /// bulk result or iterator.  A [futures::Stream](https://docs.rs/futures/0.3.3/futures/stream/trait.Stream.html)
-    /// can be obtained by calling `stream()` on the AsyncIter.  In normal mode this is not in any way more
+    /// is implemented on AsyncIter. In normal mode this is not in any way more
     /// efficient than just querying into a `Vec<T>` as it's internally
     /// implemented as buffering into a vector.  This however is useful when
     /// `cursor_arg` was used in which case the stream will query for more
@@ -434,9 +495,9 @@ impl Cmd {
         let rv = con.req_packed_command(&self).await?;
 
         let (cursor, batch) = if rv.looks_like_cursor() {
-            from_redis_value::<(u64, Vec<T>)>(&rv)?
+            from_owned_redis_value::<(u64, Vec<T>)>(rv)?
         } else {
-            (0, from_redis_value(&rv)?)
+            (0, from_owned_redis_value(rv)?)
         };
         if cursor == 0 {
             self.cursor = None;
@@ -445,9 +506,11 @@ impl Cmd {
         }
 
         Ok(AsyncIter {
-            batch: batch.into_iter(),
-            con,
-            cmd: self,
+            inner: IterOrFuture::Iter(AsyncIterInner {
+                batch: batch.into_iter(),
+                con,
+                cmd: self,
+            }),
         })
     }
 
@@ -461,15 +524,34 @@ impl Cmd {
     /// ```rust,no_run
     /// # let client = redis::Client::open("redis://127.0.0.1/").unwrap();
     /// # let mut con = client.get_connection().unwrap();
-    /// let _ : () = redis::cmd("PING").query(&mut con).unwrap();
+    /// redis::cmd("PING").query::<()>(&mut con).unwrap();
     /// ```
     #[inline]
+    #[deprecated(note = "Use Cmd::exec + unwrap, instead")]
     pub fn execute(&self, con: &mut dyn ConnectionLike) {
-        self.query::<()>(con).unwrap();
+        self.exec(con).unwrap();
+    }
+
+    /// This is an alternative to `query`` that can be used if you want to be able to handle a
+    /// command's success or failure but don't care about the command's response. For example,
+    /// this is useful for "SET" commands for which the response's content is not important.
+    /// It avoids the need to define generic bounds for ().
+    #[inline]
+    pub fn exec(&self, con: &mut dyn ConnectionLike) -> RedisResult<()> {
+        self.query::<()>(con)
+    }
+
+    /// This is an alternative to `query_async` that can be used if you want to be able to handle a
+    /// command's success or failure but don't care about the command's response. For example,
+    /// this is useful for "SET" commands for which the response's content is not important.
+    /// It avoids the need to define generic bounds for ().
+    #[cfg(feature = "aio")]
+    pub async fn exec_async(&self, con: &mut impl crate::aio::ConnectionLike) -> RedisResult<()> {
+        self.query_async::<()>(con).await
     }
 
     /// Returns an iterator over the arguments in this command (including the command name itself)
-    pub fn args_iter(&self) -> impl Iterator<Item = Arg<&[u8]>> + Clone + ExactSizeIterator {
+    pub fn args_iter(&self) -> impl Clone + ExactSizeIterator<Item = Arg<&[u8]>> {
         let mut prev = 0;
         self.args.iter().map(move |arg| match *arg {
             Arg::Simple(i) => {
@@ -506,6 +588,19 @@ impl Cmd {
         }
         Some(&self.data[start..end])
     }
+
+    /// Client won't read and wait for results. Currently only used for Pub/Sub commands in RESP3.
+    #[inline]
+    pub fn set_no_response(&mut self, nr: bool) -> &mut Cmd {
+        self.no_response = nr;
+        self
+    }
+
+    /// Check whether command's result will be waited for.
+    #[inline]
+    pub fn is_no_response(&self) -> bool {
+        self.no_response
+    }
 }
 
 /// Shortcut function to creating a command with a single argument.
